class Token:
    def __init__(self, token_type: str, value: str) -> None:
        self.type: str = token_type
        self.value: str = value

    def __repr__(self) -> str:
        return f"{self.value}"


class Integer(Token):
    def __init__(self, value: str) -> None:
        super().__init__("INT", value)


class Float(Token):
    def __init__(self, value: str) -> None:
        super().__init__("FLT", value)


class Operator(Token):
    def __init__(self, value: str) -> None:
        super().__init__("OP", value)


class Declaration(Token):
    def __init__(self, value: str) -> None:
        super().__init__("DECL", value)


class Variable(Token):
    def __init__(self, value: str) -> None:
        super().__init__("VAR(?)", value)



from .tokens import Token
from .typedef import ParseTreeType, TokenListType, VariableType


class Parser:
    def __init__(self, tokens: TokenListType) -> None:
        # type: ignore # To disable pylance warnings
        self.tokens: TokenListType = tokens
        self.idx: int = 0

        self.token: Token = self.tokens[self.idx]

    def look_ahead(self) -> None:
        self.idx += 1
        if self.idx < len(self.tokens):
            self.token = self.tokens[self.idx]

    def variable(self) -> VariableType:
        if self.token.type.startswith("VAR"):
            temp = self.token
            self.look_ahead()
            return temp

        if self.idx < len(self.tokens):
            self.token = self.tokens[self.idx]

        return None

    def factor(self) -> ParseTreeType:
        """A factor can be an int, float, expression inside parenthesis or a variable.

        Grammar:
            <factor> = INT | FLT | (<expr>) | VAR

        Returns:
            Token | List[Token]
        """

        if self.token.type in ("INT", "FLT"):
            return self.token

        elif self.token.value == "(":
            # To skip '('
            self.look_ahead()
            # To evaluate the expression in between parenthesis '(<expr>)'
            expression: ParseTreeType = self.expression()

            return expression

        elif self.token.type.startswith("VAR"):
            return self.token

        elif self.token.value in ("+", "-"):
            operator = self.token
            self.look_ahead()
            operand = self.factor()

            return [operator, operand]

        return None

    def term(self) -> ParseTreeType:
        left_node: ParseTreeType = self.factor()
        self.look_ahead()

        while self.token.value in ("*", "/"):
            operator = self.token
            self.look_ahead()

            right_node: ParseTreeType = self.factor()
            left_node = [left_node, operator, right_node]

        return left_node

    def expression(self) -> ParseTreeType:
        left_node: ParseTreeType = self.term()

        while self.token.value in ("+", "-"):
            operator: Token = self.token
            self.look_ahead()

            right_node: ParseTreeType = self.term()
            left_node = [left_node, operator, right_node]

        return left_node

    def statement(self) -> ParseTreeType:
        # Variable assignment
        if self.token.type == "DECL":
            self.look_ahead()
            left_node: VariableType = self.variable()
            if self.token.value == "=":
                operator: Token = self.token
                self.look_ahead()
                right_node: ParseTreeType = self.expression()

                return [left_node, operator, right_node]

        # Arithmetic expression
        elif self.token.type in ("INT", "FLT", "OP"):
            return self.expression()

        return None

    def parse(self) -> ParseTreeType:
        return self.statement()



from typing import cast

from utils.console import console
from utils.helpers import datatype_casting

from .memory import Memory
from .tokens import Float, Integer, Token
from .typedef import BinaryType, ParseTreeType, PrintDataType, UnaryType


class Interpreter:
    def __init__(self, tree: ParseTreeType, memory: Memory) -> None:
        if tree is None:
            self.syntax_error = True
        else:
            self.syntax_error = False

        self.tree: ParseTreeType = tree
        self.memory: Memory = memory

    @staticmethod
    def INT(value: str) -> int:
        return int(value)

    @staticmethod
    def FLT(value: str) -> float:
        return float(value)

    def VAR(self, id: str) -> UnaryType:
        variable: Token = self.memory.read(id)
        variable_type: str = variable.type

        value = getattr(self, f"{variable_type}")(variable.value)

        return datatype_casting(value, variable_type)

    def compute_unary(self, operator: Token, operand: Token) -> UnaryType:
        operand_type: str = "VAR" if operand.type.startswith("VAR") else operand.type

        operand_value = datatype_casting(getattr(self, f"{operand_type}")(operand.value), operand_type)

        match operator.value:
            case "+":
                operand_value = +operand_value
            case "-":
                operand_value = -operand_value

        return operand_value

    def compute_binary(self, left_operand: Token, operator: Token, right_operand: Token) -> BinaryType:
        left_type: str = "VAR" if left_operand.type.startswith("VAR") else left_operand.type
        right_type: str = "VAR" if left_operand.type.startswith("VAR") else right_operand.type

        if operator.value == "=":
            left_operand.type = f"VAR{right_type}"
            self.memory.write(left_operand, right_operand)
            return right_operand.value

        left_operand_value = getattr(self, f"{left_type}")(left_operand.value)
        right_operand_value = getattr(self, f"{right_type}")(right_operand.value)

        match operator.value:
            case "+":
                result = left_operand_value + right_operand_value
            case "-":
                result = left_operand_value - right_operand_value
            case "*":
                result = left_operand_value * right_operand_value
            case "/":
                result = left_operand_value / right_operand_value

        if "FLT" in (left_type, right_type):
            return Float(result)
        else:
            return Integer(result)

    def interpret(self, tree: ParseTreeType = None) -> PrintDataType:
        if self.syntax_error:
            console.print("[b][ERROR]:[/b] Syntax error", style="error")
            return None

        if tree is None:
            tree = self.tree

        # Arithmetic literal
        if not isinstance(tree, list):
            return tree

        # Unary opeartion
        elif isinstance(tree, list) and len(tree) == 2:
            # compute_unary(operator, operand)
            unary_operator = cast(Token, tree[0])
            unary_operand = cast(Token, tree[1])
            return self.compute_unary(unary_operator, unary_operand)

        # Binary operations
        else:
            left_operand = tree[0]
            if isinstance(left_operand, list):
                left_operand = cast(Token, self.interpret(left_operand))

            right_operand = tree[2]
            if isinstance(right_operand, list):
                right_operand = cast(Token, self.interpret(right_operand))

            left_operand = cast(Token, left_operand)
            right_operand = cast(Token, right_operand)
            operator = cast(Token, tree[1])

            return self.compute_binary(left_operand, operator, right_operand)



from typing import List, TypeAlias, Union

from .tokens import Float, Integer, Token

TokenListType: TypeAlias = List[Token]
NumberType: TypeAlias = Union[Integer, Float]
WordType: TypeAlias = str

VariableType: TypeAlias = Token | None
ParseTreeType: TypeAlias = Union[Token, List[Token], List["ParseTreeType"], None]

UnaryType: TypeAlias = Union[int, float]
BinaryType: TypeAlias = Union[Integer, Float, str]

PrintDataType: TypeAlias = Union[UnaryType, BinaryType, ParseTreeType]



from utils import constants

from .tokens import Declaration, Float, Integer, Operator, Token, Variable
from .typedef import NumberType, TokenListType, WordType


class Lexer:
    def __init__(self, text: str) -> None:
        self.text: str = text
        self.idx: int = 0
        self.tokens: TokenListType = []
        self.token: Token
        self.char: str = self.text[self.idx]

    def look_ahead(self) -> None:
        self.idx += 1

        if self.idx < len(self.text):
            self.char = self.text[self.idx]

    def extract_number(self) -> NumberType:
        numbers: str = ""
        isFloat: bool = False

        while (self.char in constants.DIGITS or self.char == ".") and (self.idx < len(self.text)):
            # Check if the number is floating point by checking if it contains decimal point(.)
            if self.char == ".":
                isFloat = True

            numbers += self.char
            self.look_ahead()

        if isFloat:
            return Float(numbers)
        else:
            return Integer(numbers)

    def extract_word(self) -> WordType:
        word: str = ""
        while self.char.lower() in constants.ALPHABETS and self.idx < len(self.text):
            word += self.char
            self.look_ahead()

        return word

    def tokenize(self) -> TokenListType:
        while self.idx < len(self.text):
            if self.char in constants.DIGITS:
                self.token = self.extract_number()

            elif self.char in constants.OPERATORS:
                self.token = Operator(self.char)
                self.look_ahead()

            elif self.char.lower() in constants.ALPHABETS:
                word: str = self.extract_word()

                if word in constants.DECLARATIONS:
                    self.token = Declaration(word)
                else:
                    self.token = Variable(word)

            elif self.char in constants.STOPWORDS:
                self.look_ahead()
                continue

            self.tokens.append(self.token)

        return self.tokens



from prompt_toolkit import PromptSession
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import Style
from pygments.lexers.javascript import JavascriptLexer
from rich.console import Console, Theme

rich_console_theme = Theme({"info": "blue", "warning": "yellow", "error": "bold red"})

console = Console(theme=rich_console_theme)


prompt_toolkit_style = Style.from_dict(
    {
        "info": "skyblue bold",
    }
)

session = PromptSession(lexer=PygmentsLexer(JavascriptLexer), style=prompt_toolkit_style)


def prompt():
    """
    Note: this is dirty fix for issue: https://github.com/Textualize/rich/issues/2293 in rich.console.input -> readline
    """

    message = [("class:info", "[REPL] Vadi: ")]

    return session.prompt(message, style=prompt_toolkit_style)



from typing import Dict

from .tokens import Token


class Memory:
    def __init__(self) -> None:
        self.variables: Dict[str, Token] = {}

    def read(self, id: str) -> Token:
        return self.variables[id]

    def read_all(self) -> Dict[str, Token]:
        return self.variables

    def write(self, variable: Token, expression: Token) -> None:
        variable_name: str = variable.value
        self.variables[variable_name] = expression



import sys

from rich.traceback import install
from utils.console import console, prompt

from .interpreter import Interpreter
from .lexer import Lexer
from .memory import Memory
from .parser import Parser  # type: ignore # To disable pylance warnings

install()


def main() -> None:
    memory = Memory()

    while True:
        try:
            text = prompt()
            # if the incoming sting text is empyt then skip the iteration, empty string means <enter> | ""
            if text == "":
                continue

            tokenizer = Lexer(text)
            tokens = tokenizer.tokenize()

            parser = Parser(tokens)
            tree = parser.parse()

            interpreter = Interpreter(tree, memory)
            result = interpreter.interpret()

            if result is not None:
                console.print(result)
        except KeyError:
            console.print("[b][ERROR][/b]: Syntax error", style="error")

        except (KeyboardInterrupt, EOFError):
            sys.exit()


if __name__ == "__main__":
    main()



from vadi.typedef import UnaryType


def datatype_casting(value: UnaryType, type: str) -> UnaryType:
    match type:
        case "INT":
            value = int(value)
        case "FLT":
            value = float(value)

    return value
